import base64
import json
import os
import random
import re
import time
from functools import singledispatch, wraps

import requests
from requests import models
from selenium import webdriver

# TODO
# from requests.exceptions import ConnectionError

_site_reader = {
    # "domain": ["reader", RegExp, param1, param2, ...]
    "r.binb.jp":                        ["binb2", "r.binb.jp/epm/([\w-]+)/", 1],
    "www.cmoa.jp":                      ["binb2", "www.cmoa.jp/bib/speedreader/speed.html\?cid=([\w-]+)&u0=(\d)&u1=(\d)", 1],
    "booklive.jp":                      ["binb", "booklive.jp/bviewer/s/\?cid=([\w-]*)&", 1],
    "takeshobo.co.jp":                  ["binb", "[\w-]+.takeshobo.co.jp/manga/([\w-]+)/_files[\w-]*/([\w-]+)/", 0],
    "www.comic-valkyrie.com":           ["binb", "www.comic-valkyrie.com/samplebook/([\w-]*)/", 0],
    "futabanet.jp":                     ["binb", "futabanet.jp/common/dld/zip/([\w-]*)/", 0],
    "comic-polaris.jp":                 ["binb", "comic-polaris.jp/ptdata/([\w-]*)/([\w-]*)/", 0],
    "www.shonengahosha.co.jp":          ["binb", "www.shonengahosha.co.jp/([\w-]*)/([\w-]*)/", 0],

    "comic-action.com":                 ["comic_action", "episode/([\w-]*)", 0],
    "comic-days.com":                   ["comic_action", "episode/([\w-]*)", 1],
    "comic-gardo.com":                  ["comic_action", "episode/([\w-]*)", 1],
    "comic-zenon.com":                  ["comic_action", "episode/([\w-]*)", 0],
    "comicborder.com":                  ["comic_action", "episode/([\w-]*)", 0],
    "kuragebunch.com":                  ["comic_action", "episode/([\w-]*)", 0],
    "magcomi.com":                      ["comic_action", "episode/([\w-]*)", 0],
    "pocket.shonenmagazine.com":        ["comic_action", "episode/([\w-]*)", 1],
    "shonenjumpplus.com":               ["comic_action", "episode/([\w-]*)", 1],
    "tonarinoyj.jp":                    ["comic_action", "episode/([\w-]*)", 0],
    "viewer.heros-web.com":             ["comic_action", "episode/([\w-]*)", 0],

    "viewer.comic-earthstar.jp":        ["comic_earthstar", "cid=([\w-]*)"],

    "comic-walker.com":                 ["comic_walker", 'cid=([\w-]*)'],

    # "viewer.ganganonline.com":          ["ganganonline", None],

    # "www.manga-doa.com":                ["manga_doa", None],

    # "www.sukima.me":                    ["sukima", None],

    # "www.sunday-webry.com":             ["sunday_webry", None],

    "urasunday.com":                    ["urasunday", None],

    "ganma.jp":                         ["ganma", "ganma.jp/(?:([\w-]*)/([\w-]*)|([\w-]*))"]
}


class SiteReaderLoad(object):

    __reader_reg = {}

    @staticmethod
    def readers():
        return [_site_reader[x][0] for x in _site_reader]

    @staticmethod
    def sites():
        return [x for x in _site_reader]

    @staticmethod
    def reader_name(site_name):
        return _site_reader[site_name][0] if site_name in _site_reader else None

    @staticmethod
    def get_param(site_name):
        r"""
        return [RegExp, param1, param2,...]
        """
        return _site_reader[site_name][1:]

    @staticmethod
    def register(reader_name):
        def decorator(reader_cls):
            SiteReaderLoad.__reader_reg[reader_name] = reader_cls
            return reader_cls
        return decorator
    
    @staticmethod
    def reader_cls(reader_name):
        if reader_name in SiteReaderLoad.__reader_reg:
            return SiteReaderLoad.__reader_reg[reader_name]
        else:
            return None


class ComicLinkInfo(object):
    def __init__(self, url):
        super().__init__()
        self._url = url
        self._site_name = None
        self._reader = None
        self._reader_func_entry = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = url

    @property
    def site_name(self):
        match = re.search('//(?:(.*?)/|(.*))', self._url)
        # match = re.search('[a-zA-Z]*//(?:.*\.(.*)|(.*))\..*?/', self._url)
        if not match:
            return None
        match = match.groups()
        match = match[0] if match[0] else match[1]
        if "takeshobo.co.jp" in match:
            match = "takeshobo.co.jp"
        return match

    @property
    def reader_name(self):
        return SiteReaderLoad.reader_name(self.site_name)

    @property
    def param(self):
        r"""
        return [param_regexp:list, param1, param2,...]
        """
        param = SiteReaderLoad.get_param(self.site_name)
        if param and type(param) == list and param[0] and type(param[0]) == str:
            search = re.search(param[0], self._url)
            if search:
                param[0] = [x for x in search.groups()]
        return param

    # @property
    # def Reader(self):
    #     return SiteReaderLoad.reader_cls(self.reader_name)


class ProgressBar(object):
    """Progress bar for terminal display

    """

    def __init__(self, total: int):
        """Inits ProgressBar with total"""
        super().__init__()
        self._space = 50
        self._total = total


    def show(self, current_set: int):
        a = int((current_set / self._total) * self._space)
        b = self._space - a
        text = "\r|{}{}| {:>3s}% ({:>} of {:>})".format(
            '#' * a, ' ' * b, str(int((current_set / self._total) * 100)), str(current_set), str(self._total))
        print(text, end='')
        if a == self._space:
            print('')


def draw_image(img_source, img_target, src_x, src_y, swidth, sheight, x, y, width=None, height=None):
    img_target.paste(
        img_source.crop
        (
            (src_x, src_y, src_x + swidth, src_y + sheight),
        ),
        (x, y))


def cc_mkdir(fpath, model=0):
    r"""
    :param model: model = 0, include two subfolders of source and target; model = 1, not include.
    """
    if model == 1:
        if os.path.exists(fpath):
            print('\n當前一話的文件夾{}存在，繼續運行數據將被覆蓋！'.format(fpath))
            print('是否繼續運行？（y/n）')
            yn = input()
            return 0 if yn == 'y' or yn == 'yes' or yn == 'Y' else -1
        else:
            if not os.path.exists(fpath):
                os.makedirs(fpath)
            print('創建文件夾: ' + fpath)
            return 0
    if os.path.exists(fpath+'/source') and os.path.exists(fpath+'/target'):
        print('\n當前一話的文件夾{}存在，繼續運行數據將被覆蓋，'.format(fpath))
        print('是否繼續運行？（y/n）')
        yn = input()
        return 0 if yn == 'y' or yn == 'yes' or yn == 'Y' else -1
    else:
        if not os.path.exists(fpath + '/source'):
            os.makedirs(fpath + '/source')
        if not os.path.exists(fpath + '/target'):
            os.makedirs(fpath + '/target')

        print('創建文件夾: ' + fpath)
        return 0

def get_blob_content(driver:webdriver.Chrome, uri):
    """
    获取浏览器中的blob对象的数据
    """
    result = driver.execute_async_script("""
        var uri = arguments[0];
        var callback = arguments[1];
        var toBase64 = function(buffer){for(var r,n=new Uint8Array(buffer),t=n.length,a=new Uint8Array(4*Math.ceil(t/3)),i=new Uint8Array(64),o=0,c=0;64>c;++c)i[c]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".charCodeAt(c);for(c=0;t-t%3>c;c+=3,o+=4)r=n[c]<<16|n[c+1]<<8|n[c+2],a[o]=i[r>>18],a[o+1]=i[r>>12&63],a[o+2]=i[r>>6&63],a[o+3]=i[63&r];return t%3===1?(r=n[t-1],a[o]=i[r>>2],a[o+1]=i[r<<4&63],a[o+2]=61,a[o+3]=61):t%3===2&&(r=(n[t-2]<<8)+n[t-1],a[o]=i[r>>10],a[o+1]=i[r>>4&63],a[o+2]=i[r<<2&63],a[o+3]=61),new TextDecoder("ascii").decode(a)};
        var xhr = new XMLHttpRequest();
        xhr.responseType = 'arraybuffer';
        xhr.onload = function(){ callback(toBase64(xhr.response)) };
        xhr.onerror = function(){ callback(xhr.status) };
        xhr.open('GET', uri);
        xhr.send();
        """, uri)
    if type(result) == int:
        raise Exception("Request failed with status %s" % result)
    return base64.b64decode(result)
    
class RqHeaders(dict):
    def __init__(self):
        super().__init__()
        self.__setitem__(
            'User-Agent', 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36')

    def random_ua(self):
        # self.__setitem__('User-Agent', )
        pass



import base64
from copy import deepcopy
import json
import logging
import math
import random
import re
import time
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO

import numpy as np
import requests
from lxml.html import etree
from PIL import Image
from requests import cookies
from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait

from .utils import (ComicLinkInfo, ProgressBar, RqHeaders, SiteReaderLoad, cc_mkdir,
                    draw_image, get_blob_content)

logger = logging.getLogger(__name__)

# 页面元素加载的最大等待时间
WAIT_TIME = 90


class N21(object):
    """
    docstring
    """

    def __init__(self, dirpath) -> None:
        super().__init__()
        if dirpath[:-1] in ('\\', '/'):
            self._dirpath = dirpath
        else:
            self._dirpath = dirpath + '/'
        try:
            os.makedirs(self._dirpath + 'target/')
        except Exception:
            pass

    def load_imgs(self, dirpath) -> dict:
        """
        :returns: i
        """
        ...
        # fl = os.listdir(dirpath)
        # fl = [x for x in fl if not os.path.isdir(dirpath + x)]
        # if len(fl) < 1:
        #     raise ValueError('No image found :{}'.format(dirpath))
        # fmt = fl[0].split('.')[-1]
        # img_dict = {}
        # for fpath in fl:
        #     f_re = [int(x)-1 for x in fpath[:-(len(fmt)+1)].split("-")]
        #     if f_re[0] not in img_dict:
        #         img_dict[f_re[0]] = [None for x in range(3)]
        #     img_dict[f_re[0]][f_re[1]] = Image.open(
        #         dirpath+fpath).convert('1')
        # for x in iter(img_dict.keys()):
        #     if None in img_dict[x]:
        #         print("page{}")
        # return img_dict

    def edge_connection_offset(self, imgs_split_3_pages):
        """
        :param imgs_split_3_pages: [[img0_0, img0_1, img_0_2], ...]
        :return: (offset_1, offset_2)
        :rtype: tuple
        """
        def calculate_similarity(a, b):
            count = 0
            img_xor = np.logical_xor(a, b)
            for z in img_xor:
                for x in z:
                    if not x:
                        count += 1
                    else:
                        count -= 1
            return count / img_xor.size

        ij = [[], []]
        for page in imgs_split_3_pages:
            for x in range(2):
                img_ = []
                for offset in range(3, 18):
                    img_.append(np.array(page[x].crop(
                        (0, page[x].height - offset, page[x].width, page[x].height - offset + 3)).convert('1')))
                b = np.array(page[x+1].crop((0, 0, page[1].width, 3)).convert('1'))
                score = [(i+3, calculate_similarity(img_[i], b))
                         for i in range(len(img_))]
                score.sort(key=lambda x: x[1], reverse=True)
                ij[x].append(score[0][0])

        offset = [None, None]
        for i in range(2):
            max_count = 0
            offset_1 = None
            for x in list(set(ij[i])):
                if max_count < ij[i].count(x):
                    max_count = ij[i].count(x)
                    offset[i] = x

        return tuple(offset)

    def crop_paste(self, img_chunks, i, j):
        """
        :param img_chunks: [img_c0, img_c1, img0_c2],
        :returns: An ~PIL.Image.Image object.
        """

        img_new = Image.new('RGB', (img_chunks[0].width, (
            img_chunks[0].height + img_chunks[1].height + img_chunks[2].height - i - j)))
        img_new.paste(img_chunks[0], (0, 0))
        img_new.paste(img_chunks[1], (0, img_chunks[0].height - i))
        img_new.paste(
            img_chunks[2], (0, img_chunks[0].height - i + img_chunks[1].height - j))
        return img_new
        # img_new.save(file_path+'/target/{}.png'.format(count))
        # print("完成!")

    def run(self, imgs, index):
        # imgs_dict = self.load_imgs(self._dirpath + "source/")
        img_new = self.crop_paste(
            imgs, *self.edge_connection_offset([imgs]))
        img_new.save(self._dirpath + 'target/{}.png'.format(index))

def gen_file_path(link_info: ComicLinkInfo, driver: webdriver.Chrome):
    if link_info.site_name == "www.cmoa.jp":
        try:
            elem = WebDriverWait(driver, WAIT_TIME, 0.5).until(
                lambda x: x.find_element_by_xpath("/html/head/title"),
                message="無法定位元素 " + "/html/head/title")
        except TimeoutException as e:
            logger.error(
                "Find element by xpath(\"/html/head/title\"): " + str(e))
            raise e
        match = re.search("[\w]+_[\w]+_([\w]+)", link_info.param[0][0])
        if match:
            return elem.get_attribute("innerText") + '/' + match.groups()[0]
        else:
            logger.error("url:{}\ncid: {}".format(link_info.url,
                                                  link_info.param[0][0]))
            raise ValueError("Unusual cid!")

    elif link_info.site_name == "r.binb.jp":
        return 'binb/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "booklive.jp":
        return 'booklive/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "takeshobo.co.jp":
        return 'takeshobo/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "www.comic-valkyrie.com":
        return 'comic-valkyrie/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "futabanet.jp":
        return 'futabanet/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "comic-polaris.jp":
        return 'comic-polaris/{}'.format(int(time.time()*1000))

    elif link_info.site_name == "www.shonengahosha.co.jp":
        return 'shonengahosha/{}'.format(int(time.time()*1000))
    # elif domain == "":
    #     pass


def getRandomString(t, i=None):
    n = i or "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"
    r = len(n)
    e = ""
    for x in range(t):
        e += n[math.floor(random.random() * r)]
    return e


def get_cookies_thr_browser(url, driver: webdriver.Chrome):
    r"""
    :param url: comic url
    :param driver: webdriver.Chrome
    :return: dict or None
    """
    for hdl in driver.window_handles:
        driver.switch_to_window(hdl)
        if driver.current_url == url:
            # return driver.get_cookies
            cookies = requests.cookies.RequestsCookieJar()
            for cookie in driver.get_cookies():
                cookies.set(cookie["name"], cookie["value"])
            return cookies
    return None


class DownldGenBinb2(object):
    def __init__(self, manga_info, link_info, base_path, cnt_p, cid, cntserver, ctbl, ptbl, content_data, u0, u1):
        self._manga_info = manga_info
        self._base_path = base_path
        self._link_info = link_info
        self._contents_server = cntserver
        self._cnt_p = cnt_p
        self._cid = cid
        self._ctbl = ctbl
        self._ptbl = ptbl
        self._content_data = content_data
        self._u0 = u0
        self._u1 = u1

    @property
    def file_path_g(self):
        for x in self._manga_info:
            if self._link_info.site_name == "www.cmoa.jp":
                yield self._base_path + "/{}.png".format(int(x["id"][1:])+1)
            elif self._link_info.site_name == "r.binb.jp":
                yield self._base_path + "/{}.png".format(int(x["id"][1:])+1)
            else:
                raise ValueError("")

    @property
    def img_url_g(self):
        for x in self._manga_info:
            if self._link_info.site_name == "www.cmoa.jp":
                yield self._contents_server + "/sbcGetImg.php?cid={}&src={}&p={}&q=1&vm=1&dmytime={}&u0={}&u1={}".format(
                    self._cid, x["src"].replace("/", "%2F"), self._cnt_p, self._content_data, self._u0, self._u1)
            elif self._link_info.site_name == "r.binb.jp":
                yield "{}{}sbcGetImg.php?cid={}&src={}&p={}&q=1".format(
                    self._link_info._url, self._contents_server, self._cid, x["src"], self._cnt_p
                )
            else:
                raise ValueError("")

    @property
    def coords(self):
        for x in self._manga_info:
            yield ImageDescrambleCoords.ctbl_ptbl(x["src"], self._ctbl, self._ptbl)


class ImageDescrambleCoords(dict):
    def __init__(self, width, height, h, s):
        r"""
        :param width: image width 
        :param height: image height
        :param h: ctbl
        :param s: ptbl
        """
        if "=" == h[0] and "=" == s[0]:
            self.lt_f(h, s)
        e = self.lt_dt({"width": width, "height": height})
        super().__init__({
            "width": e["width"],
            "height": e["height"],
            "transfers": [{
                "index": 0,
                "coords": self.lt_bt({"width": width, "height": height})
            }],
        })

    @staticmethod
    def ctbl_ptbl(t, ctbl_d, ptbl_d):
        r"""
        :param t: t="pages/BouPrlub.jpg"
        :param ctbl_d: ctbl(decoded), list
        :param ptbl_d: ptbl(decoded), list
        """
        i = [0, 0]
        if(t):
            n = t.index('/') + 1
            r = len(t) - n
            for e in range(r):
                i[e % 2] += ord(t[e+n])
            i[0] %= 8
            i[1] %= 8
        h = ctbl_d[i[1]]
        s = ptbl_d[i[0]]
        return h, s

    def lt_f(self, t, i):
        r"""
        :param t: ctbl,str 
        :param i: ptbl,str
        """
        # "^=([0-9]+)-([0-9]+)([-+])([0-9]+)-([-_0-9A-Za-z]+)$/"
        n = re.match(
            "=([0-9]+)-([0-9]+)([-+])([0-9]+)-([-_0-9A-Za-z]+)", t).groups()
        r = re.match(
            "=([0-9]+)-([0-9]+)([-+])([0-9]+)-([-_0-9A-Za-z]+)", i).groups()
        self._lt_T = int(n[0], 10)
        self._lt_j = int(n[1], 10)
        self._lt_xt = int(n[3], 10)
        e = self._lt_T + self._lt_j + self._lt_T * self._lt_j
        if(len(n[4]) == e and len(r[4]) == e):
            s = self.lt_St(n[4])
            h = self.lt_St(r[4])
            self._lt_Ct = s["n"]
            self._lt_At = s["t"]
            self._lt_Tt = h["n"]
            self._lt_Pt = h["t"]
            self._lt_It = []
            for x in range(self._lt_T * self._lt_j):
                self._lt_It.append(s["p"][h["p"][x]])

    def lt_bt(self, t):
        r"""
        :param t: {"width": img width, "height": img height}
        """
        i = t["width"] - 2 * self._lt_T * self._lt_xt
        n = t["height"] - 2 * self._lt_j * self._lt_xt
        r = math.floor((i + self._lt_T - 1) / self._lt_T)
        e = i - (self._lt_T - 1) * r
        s = math.floor((n + self._lt_j - 1) / self._lt_j)
        h = n - (self._lt_j - 1) * s
        u = []
        for o in range(self._lt_T * self._lt_j):
            a = o % self._lt_T
            f = math.floor(o / self._lt_T)
            c = self._lt_xt + a * (r + 2 * self._lt_xt) + \
                (e - r if self._lt_Tt[f] < a else 0)
            l = self._lt_xt + f * (s + 2 * self._lt_xt) + \
                (h - s if self._lt_Pt[a] < f else 0)
            v = self._lt_It[o] % self._lt_T
            d = math.floor(self._lt_It[o] / self._lt_T)
            b = v * r + (e - r if self._lt_Ct[d] < v else 0)
            g = d * s + (h - s if self._lt_At[v] < d else 0)
            p = e if self._lt_Tt[f] == a else r
            m = h if self._lt_Pt[a] == f else s
            0 < i and 0 < n and u.append({
                "xsrc": c,
                "ysrc": l,
                "width": p,
                "height": m,
                "xdest": b,
                "ydest": g
            })
        return u

    def lt_dt(self, t):
        r"""
        :param t: t={width: int, height: int}
        """
        i = 2 * self._lt_T * self._lt_xt
        n = 2 * self._lt_j * self._lt_xt
        if t["width"] >= 64 + i and t["height"] >= 64 + n and t["width"] * t["height"] >= (320 + i) * (320 + n):
            return {
                "width": t["width"] - 2 * self._lt_T * self._lt_xt,
                "height": t["height"] - 2 * self._lt_j * self._lt_xt
            }
        else:
            return t

    def lt_St(self, t):
        kt = [-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -
              1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, 63, -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1]
        n = []
        r = []
        e = []
        for x in range(self._lt_T):
            n.append(kt[ord(t[x])])

        for x in range(self._lt_j):
            r.append(kt[ord(t[self._lt_T + x])])

        for x in range(self._lt_T * self._lt_j):
            e.append(kt[ord(t[self._lt_T + self._lt_j + x])])

        # t n p
        return {"t": n, "n": r, "p": e}


@SiteReaderLoad.register("binb")
class Binb(object):
    def __init__(self, link_info: ComicLinkInfo, driver: webdriver.Chrome):
        super().__init__()
        self._link_info = link_info
        self._driver = driver

    def page_number(self):
        """
        return [current_page_numb:int, total_numb:int]
        """
        page_elem_id = 'menu_slidercaption'
        match = re.search("cid=([0-9a-zA-Z_]+)", self._driver.current_url)
        pageNum = []
        count = 8
        while len(pageNum) != 2:
            if count > 0:
                count -= 1
                time.sleep(0.35)
            else:
                raise TimeoutException("")
            try:
                elemt = WebDriverWait(self._driver, WAIT_TIME, 0.5).until(
                    lambda x: x.find_element_by_id(page_elem_id),
                    message="無法定位元素 " + page_elem_id
                )
            except TimeoutException as e:
                logger.error(str(e))
                raise e
            pageNum = elemt.get_attribute('innerText').split('/')
        pageNum[1] = int(pageNum[1])
        if (pageNum[0] != '-'):
            pageNum[0] = int(pageNum[0])
        else:
            pageNum[0] = pageNum[1]
        return pageNum

    def downloader(self):
        # self._driver.get(self._link_info.url)
        self._driver.get(self._link_info.url)
        self.page_number()
        file_path = "./漫畫/" + gen_file_path(self._link_info, self._driver)
        if cc_mkdir(file_path) != 0:
            return -1
        n21 = N21(file_path)
        while(self.page_number()[0] > 1):
            ActionChains(self._driver).send_keys(Keys.RIGHT).perform()
        progress_bar = ProgressBar(self.page_number()[1])
        start_page = 0
        reader_flag = self._link_info.param[1]
        while(start_page < self.page_number()[1]):
            for i in range(self.page_number()[0], start_page, -1):
                imgs = []
                for j in range(1, 4):
                    try:
                        blob_url = WebDriverWait(self._driver, WAIT_TIME, 0.5).until(
                            lambda x: x.find_element_by_xpath(
                                '//*[@id="content-p{}"]/div/div[{}]/img'.format(i-(1-reader_flag), j)),
                            message="p" + str(i) + "part" + str(j) + "無法定位"
                        )
                    except TimeoutException as e:
                        logger.error(
                            "Find element by id(\"/html/head/title\"): " + str(e))
                        raise e
                    blob_uri = blob_url.get_attribute('src')
                    try:
                        img = Image.open(
                            BytesIO(get_blob_content(self._driver, blob_uri)))
                        img.save(
                            file_path+'/source/{}-{}.png'.format(i, j))
                        imgs.append(img)
                    except Exception as e:
                        logger.error(str(e))
                        raise e
                n21.run(imgs, i)
                start_page += 1
                progress_bar.show(start_page)
            ActionChains(self._driver).send_keys(Keys.LEFT).perform()


@SiteReaderLoad.register('binb2')
class Binb2(object):
    def __init__(self, link_info: ComicLinkInfo, driver: webdriver.Chrome):
        super().__init__()
        self._link_info = link_info
        self._webdriver = driver
        self._url = link_info.url
        self._header = RqHeaders()

    def gen_cntntInfo_url(self, **kwargs):
        if self._link_info.site_name == "www.cmoa.jp":
            return "https://www.cmoa.jp/bib/sws/bibGetCntntInfo.php?cid={}&dmytime={}&k={}&u0={}&u1={}".format(
                self._cid, int(time.time()*1000), self._rq_k, self._u0, self._u1)
        elif self._link_info.site_name == "r.binb.jp":
            return "{}swsapi/bibGetCntntInfo?cid={}&dmytime={}&k={}".format(
                self._link_info.url, self._cid, int(time.time()*1000), self._rq_k)

    def gen_GetCntnt_url(self, **kwargs):
        if self._link_info.site_name == "www.cmoa.jp":
            # print("https://binb-cmoa.sslcs.cdngc.net/sbc/sbcGetCntnt.php?cid={}&p={}&vm=1&dmytime={}&u0=0&u1=0".format(
            #     self._cid, self._cnt_p, self._content_date))
            return "{}/sbcGetCntnt.php?cid={}&p={}&vm=1&dmytime={}&u0={}&u1={}".format(
                self._contents_server, self._cid, self._cnt_p, self._content_date, self._u0, self._u1)
        elif self._link_info.site_name == "r.binb.jp":
            return self._link_info.url + "{}sbcGetCntnt.php?cid={}&p={}&dmytime={}".format(
                self._contents_server, self._cid, self._cnt_p, int(time.time()*1000))

    def gen_image_url(self, img_src, **kwargs):
        if self._link_info.site_name == "www.cmoa.jp":
            return self._contents_server + "/sbcGetImg.php?cid={}&src={}&p={}&q=1&dmytime={}".format(
                self._cid, img_src, self._cnt_p, self._content_date)
        elif self._link_info.site_name == "r.binb.jp":
            return "{}{}sbcGetImg.php?cid={}&src={}&p={}&q=1".format(
                self._url, self._contents_server, self._cid, img_src, self._cnt_p
            )

    def genK(self):
        t = self._cid
        n = getRandomString(16)
        i = t
        r = i[:16]
        e = i[-16:]
        s = 0
        h = 0
        u = 0
        k = []
        for i in range(len(n)):
            s ^= ord(n[i])
            h ^= ord(r[i])
            u ^= ord(e[i])
            k.append(
                n[i] + "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"[s+h+u & 63]
            )
        self._rq_k = "".join(k)
        return self._rq_k

    def CntntInfoDecode(self, t, i, n):
        r"""
        :param t: cid
        :param i: k
        :param n: ctbl ...
        """
        r = t + ":" + i
        e = 0
        for s in range(len(r)):
            e += ord(r[s]) << s % 16
        e &= 2147483647
        if e == 0:
            e = 305419896
        h = ""
        u = e
        for s in range(len(n)):
            if u < 0:
                u = u & 0xffffffff >> 1 ^ 1210056708 & -(1 & u)
            else:
                u = u >> 1 ^ 1210056708 & -(1 & u)
            o = (ord(n[s]) - 32 + u) % 94 + 32
            h += chr(o)
        try:
            return json.loads(h)
        except Exception as e:
            return None

    # step 1
    def get_access(self):
        r"""
        :return: cid, rq.cookies, k
        """
        rq = requests.get(self._link_info.url)
        self._u0 = None
        self._u1 = None
        if rq.status_code != 200:
            e_str = "status_code:{} url:{}".format(
                rq.status_code, self._link_info.url)
            logging.error(e_str)
            raise ValueError(e_str)

        self._cookies = rq.cookies

        if self._link_info.site_name == "r.binb.jp":
            cid = re.search("data-ptbinb-cid=\"(.*?)\"", rq.text).groups()[0]
        elif self._link_info.site_name == "www.cmoa.jp":
            cid = self._link_info.param[0][0]
            self._u0 = self._link_info.param[0][1]
            self._u1 = self._link_info.param[0][2]
            if self._u0 == "0":
                self._cookies = get_cookies_thr_browser(
                    self._url, self._webdriver)

        else:
            logger.error("cid not found, url:{}".format(self._link_info.url))
            raise ValueError("cid not found")
        self._cid = cid

        return cid, self._cookies, self.genK()

    # step 2_1
    def bibGetCntntInfo(self):
        r"""

        """
        rq = requests.get(self.gen_cntntInfo_url(), cookies=self._cookies)
        if rq.status_code != 200:
            return -1
        cntntinfo = rq.json()
        self._manga_title = cntntinfo["items"][0]["Title"]

        if not self._manga_title:
            self._manga_title = None
        else:
            # Windows 路径 非法字符
            self._manga_title = re.sub(
                "[\|\*\<\>\"\\\/\:]", "_", self._manga_title)
            self._manga_title = self._manga_title.replace('?', '？')

        self._manga_subtitle = cntntinfo["items"][0]["SubTitle"]
        if not self._manga_subtitle:
            self._manga_subtitle = None
        else:
            self._manga_subtitle = re.sub(
                "[\|\*\<\>\"\\\/\:]", "_", self._manga_subtitle)
            self._manga_subtitle = self._manga_subtitle.replace('?', '？')

        if not (self._manga_subtitle or self._manga_title):
            raise ValueError("")

        self._ptbl = cntntinfo["items"][0]["ptbl"]
        self._ctbl = cntntinfo["items"][0]["ctbl"]
        self._cnt_p = cntntinfo["items"][0]["p"]
        if "ContentDate" in cntntinfo["items"][0]:
            self._content_date = cntntinfo["items"][0]["ContentDate"]
        else:
            self._content_date = None
        self._contents_server = cntntinfo["items"][0]["ContentsServer"]
        self._ptbl = self.CntntInfoDecode(self._cid, self._rq_k, self._ptbl)
        self._ctbl = self.CntntInfoDecode(self._cid, self._rq_k, self._ctbl)

    # step 2_2
    def sbcGetCntnt(self,):
        rq = requests.get(self.gen_GetCntnt_url(), cookies=self._cookies)
        if rq.status_code != 200:
            return -1
        html_manga = rq.json()["ttx"]
        manga_info = re.findall('<t-img.*?id="L[\d]+".*?>', html_manga)
        page = etree.HTML(html_manga)
        manga_info_list = [re.search(
            'src="(.*?)".*orgwidth="([\d]+)".*orgheight="([\d]+)".*id="(L[\d]+)"', x).groups() for x in manga_info]
        self._manga_info = [
            {"id": x[3], "src":x[0], "orgwidth":x[1], "orgheight":x[2]} for x in manga_info_list]
        return self._manga_info

    # step_3
    @staticmethod
    def downld_one(url, fpath, hs, cookies):
        rq = requests.get(url=url, cookies=cookies, headers=RqHeaders())
        if rq.status_code != 200:
            raise ValueError("error:down_one:" + url)
        img = Image.open(BytesIO(rq.content))
        coords = ImageDescrambleCoords(img.width, img.height, hs[0], hs[1])
        img_target = deepcopy(img)
        img_target = Image.new(img.mode, (coords["width"], coords["height"]))
        for y in coords["transfers"][0]["coords"]:
            draw_image(img, img_target, y["xsrc"], y["ysrc"],
                       y["width"], y["height"], y["xdest"], y["ydest"])
        img_target.save(fpath)
        return 0

    def downloader(self):
        self.get_access()
        self.bibGetCntntInfo()
        self.sbcGetCntnt()
        if self._manga_title and self._manga_subtitle:
            base_path = "./漫畫/" + \
                "/".join((self._manga_title, self._manga_subtitle))
        else:
            base_path = "./漫畫/" + self._manga_title if self._manga_title else self._manga_subtitle

        if cc_mkdir(base_path, model=1) != 0:
            return -1
        progressBar = ProgressBar(len(self._manga_info))
        downGen = DownldGenBinb2(self._manga_info, self._link_info, base_path, self._cnt_p,
                                 self._cid, self._contents_server, self._ctbl, self._ptbl, self._content_date, self._u0, self._u1)
        count = 0
        progressBar.show(count)
        with ThreadPoolExecutor(max_workers=4) as executor:
            for x in executor.map(self.downld_one, downGen.img_url_g, downGen.file_path_g, downGen.coords, [self._cookies for x in range(len(self._manga_info))]):
                count += 1
                progressBar.show(count)



import datetime
import logging
import re
from concurrent.futures import ThreadPoolExecutor

import requests
from requests.models import Response

from .utils import (ComicLinkInfo, ProgressBar, RqHeaders, SiteReaderLoad,
                    cc_mkdir)

logger = logging.getLogger(__name__)


@SiteReaderLoad.register("urasunday")
class Urasunday(object):
    def __init__(self, linkinfo: ComicLinkInfo, driver=None) -> None:
        super().__init__()
        self._linkinfo = linkinfo

    def page_info(self):
        resp = requests.get(url=self._linkinfo._url, headers=RqHeaders())
        Response.raise_for_status(resp)
        html_text = resp.text

        manga_title = re.search(
            '<meta property="og:title" content="(.*?)" />', html_text)
        if manga_title:
            manga_title = manga_title.groups()[0]
        else:
            manga_title = str(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"))

        img_url_l_str = re.search("const pages = ([\S\s]+?);", html_text)
        if not img_url_l_str:
            raise ValueError("Pages not found")
        img_url_l_str = img_url_l_str.groups()[0]
        src_group = re.findall('(http.+?)(?:\'|\")', img_url_l_str)
        if not src_group:
            raise ValueError("empty url list")

        return src_group, manga_title

    @staticmethod
    def downld_one(url, fpath):
        rq = requests.get("".join(url))
        if rq.status_code != 200:
            raise ValueError("".join(url))
        with open(fpath, 'wb') as fp:
            fp.write(rq.content)

    def downloader(self):
        base_file_path = './漫畫/'
        pages, manga_title = self.page_info()
        base_file_path += manga_title + '/'
        if cc_mkdir(base_file_path, 1) != 0:
            return -1
        show_bar = ProgressBar(len(pages))
        with ThreadPoolExecutor(max_workers=6) as executor:
            count = 0
            for x in executor.map(self.downld_one, pages, [base_file_path + str(x) + '.jpg' for x in range(len(pages))]):
                count += 1
                show_bar.show(count)




from .comic_action import ComicAction
from .comic_earthstar import ComicEarthstar
from .comic_walker import ComicWalker
from .ganma import Ganma
from .utils import ComicLinkInfo
from .binb import Binb, Binb2
from .urasunday import Urasunday

# from . import binb
# from . import comic_earthstar
# from . import comic_walker
# from . import ganganonline
# from . import manga_doa
# from . import sukima
# from . import sunday_webry


__all__ = [
    'ComicLinkInfo', 'ComicAction', 'ComicEarthstar', 'ComicWalker', 'Ganma', 'Binb', 'Binb2', 'Urasunday'
]



import logging
import math
import os
from os import path
import re
from copy import deepcopy
from io import BytesIO
from concurrent.futures import ThreadPoolExecutor

import execjs
import requests
from PIL import Image

from .utils import ComicLinkInfo, ProgressBar, RqHeaders, SiteReaderLoad, cc_mkdir, draw_image

logger = logging.getLogger(__name__)


js_code = '''
a3E = {
    a3f: function(a, f, b, e, d) {
        var c = Math.floor(a / b),
            g = Math.floor(f / e);
        a %= b;
        f %= e;
        var h, l, k, m, p, r, t, q, v = [];
        h = c - 43 * d % c;
        h = 0 == h % c ? (c - 4) % c : h;
        h = 0 == h ? c - 1 : h;
        l = g - 47 * d % g;
        l = 0 == l % g ? (g - 4) % g : l;
        l = 0 == l ? g - 1 : l;
        0 < a && 0 < f && (k = h * b,
            m = l * e,
            v.push({
                srcX: k,
                srcY: m,
                destX: k,
                destY: m,
                width: a,
                height: f
            }));
        if (0 < f)
            for (t = 0; t < c; t++)
                p = a3E.calcXCoordinateXRest_(t, c, d),
                k = a3E.calcYCoordinateXRest_(p, h, l, g, d),
                p = a3E.calcPositionWithRest_(p, h, a, b),
                r = k * e,
                k = a3E.calcPositionWithRest_(t, h, a, b),
                m = l * e,
                v.push({
                    srcX: k,
                    srcY: m,
                    destX: p,
                    destY: r,
                    width: b,
                    height: f
                });
        if (0 < a)
            for (q = 0; q < g; q++)
                k = a3E.calcYCoordinateYRest_(q, g, d),
                p = a3E.calcXCoordinateYRest_(k, h, l, c, d),
                p *= b,
                r = a3E.calcPositionWithRest_(k, l, f, e),
                k = h * b,
                m = a3E.calcPositionWithRest_(q, l, f, e),
                v.push({
                    srcX: k,
                    srcY: m,
                    destX: p,
                    destY: r,
                    width: a,
                    height: e
                });
        for (t = 0; t < c; t++)
            for (q = 0; q < g; q++)
                p = (t + 29 * d + 31 * q) % c,
                k = (q + 37 * d + 41 * p) % g,
                r = p >= a3E.calcXCoordinateYRest_(k, h, l, c, d) ? a : 0,
                m = k >= a3E.calcYCoordinateXRest_(p, h, l, g, d) ? f : 0,
                p = p * b + r,
                r = k * e + m,
                k = t * b + (t >= h ? a : 0),
                m = q * e + (q >= l ? f : 0),
                v.push({
                    srcX: k,
                    srcY: m,
                    destX: p,
                    destY: r,
                    width: b,
                    height: e
                });
        return v
    },
    calcPositionWithRest_: function(a, f, b, e) {
        return a * e + (a >= f ? b : 0)
    },
    calcXCoordinateXRest_: function(a, f, b) {
        return (a + 61 * b) % f
    },
    calcYCoordinateXRest_: function(a, f, b, e, d) {
        var c = 1 === d % 2;
        (a < f ? c : !c) ? (e = b,
            f = 0) : (e -= b,
            f = b);
        return (a + 53 * d + 59 * b) % e + f
    },
    calcXCoordinateYRest_: function(a, f, b, e, d) {
        var c = 1 == d % 2;
        (a < b ? c : !c) ? (e -= f,
            b = f) : (e = f,
            b = 0);
        return (a + 67 * d + f + 71) % e + b
    },
    calcYCoordinateYRest_: function(a, f, b) {
        return (a + 73 * b) % f
    }
}
'''

run_js = execjs.compile(js_code)

def setArrayPosi(width, height, num):     
    return run_js.call('a3E.a3f', width, height, 64, 64, num)

def pattern(strs):
    u = 0
    for x in strs+'/0':
        u += ord(x)
    return u % 4 + 1

class DownldGen(object):
    def __init__(self, contents, base_fpath, base_url):
        super().__init__()
        self._contents = contents
        self._base_fpath = base_fpath
        self._base_url = base_url
    
    @property
    def file_path_g(self):
        for x in self._contents:
            yield [self._base_fpath, re.search("item/xhtml/([\w-]+).xhtml", x["file"]).group(1) + ".png"]

    @property
    def img_url_g(self):
        for x in self._contents:
            yield [self._base_url, x['file'], '/0.jpeg']
    

@SiteReaderLoad.register('comic_earthstar')
class ComicEarthstar(object):
    def __init__(self, link_info: ComicLinkInfo, driver = None):
        super().__init__()
        self._link_info = link_info
        self._driver = driver
    
    @staticmethod
    def downld_one(url: list, fpath: list):
        rq = requests.get("".join(url))
        if rq.status_code != 200:
            raise ValueError("".join(url))
        img = Image.open(BytesIO(rq.content))
        img_t = deepcopy(img)
        arrayP = setArrayPosi(img.width, img.height, pattern(url[1]))
        for e in arrayP:
            draw_image(img, img_t, e['destX'], e['destY'], e['width'], e['height'],
                        e['srcX'], e['srcY'])
        img.save(fpath[0] + '/source/' + fpath[1])
        img_t.save(fpath[0] + '/target/' + fpath[1])

    def downloader(self):
        cid = self._link_info.param[0][0]

        comic_info = requests.get(headers=RqHeaders(),
                                    url='https://api.comic-earthstar.jp/c.php?cid='+cid).json()
        base_file_path = './漫畫/' + re.search(
                'https://storage.comic-earthstar.jp/data/([0-9a-zA-Z]*)/[0-9a-zA-Z_-]*/', comic_info['url']).group(1) + "/"+ comic_info["cti"]
        if cc_mkdir(base_file_path) != 0:
            return -1
        configuration = requests.get(
            headers=RqHeaders(), url=comic_info['url'] + 'configuration_pack.json')
        show_bar = ProgressBar(len(configuration.json()['configuration']['contents']))
        downldGen = DownldGen(configuration.json()['configuration']['contents'], base_file_path, comic_info['url'])
        with ThreadPoolExecutor(max_workers=4) as executor:
            count = 0
            for x in executor.map(self.downld_one,
                                  downldGen.img_url_g,
                                  downldGen.file_path_g):
                count += 1
                show_bar.show(count)
        # https://viewer.comic-earthstar.jp/viewer.html?cid=59e0b2658e9f2e77f8d4d83f8d07ca84&cty=1&lin=0







import json
import logging
from concurrent.futures import ThreadPoolExecutor
from operator import add
from os import write
from time import sleep, time

import requests
from requests.api import get
from selenium import webdriver

from .utils import ComicLinkInfo, ProgressBar, RqHeaders, cc_mkdir, SiteReaderLoad

logger = logging.getLogger(__name__)


class DownldGen(object):
    def __init__(self, item):
        self._item = item

    @property
    def file_path_g(self):
        base_path = "./漫畫/" + \
            "/".join((self._item["series"]["title"], self._item["title"]))
        for x in self._item["page"]["files"]:
            yield base_path + "/" + x

    @property
    def img_url_g(self):
        base_url = self._item["page"]["baseUrl"]
        token = self._item["page"]["token"]
        for x in self._item["page"]["files"]:
            yield base_url + x + token


class GanmaRqHeaders(RqHeaders):
    def __init__(self, manga_alias=None):
        super().__init__()
        self._manga_alias = manga_alias

    def aslogin(self):
        self.__setitem__("x-from", "https://ganma.jp/_cd/login")
        return self

    def aslogout(self):
        self.__setitem__(
            "x-from",
            "https://ganma.jp/announcement",
        )
        return self

    def asmangainfo(self):
        self.__setitem__("x-from", "https://ganma.jp/" + self._manga_alias)
        return self


@SiteReaderLoad.register("ganma")
class Ganma(object):
    def __init__(self, link_info: ComicLinkInfo, driver=None):
        super().__init__()
        self._link_info = link_info
        self._param = self._link_info.param[0]
        if self._param[2]:
            self._param[0], self._param[2] = self._param[2], self._param[0]
        elif not self._param[0] and not self._param[2]:
            raise ValueError(
                "Unable to get comic alias from URL! URL:{}".format(
                    self._link_info.url))
        self._manga_alias = self._param[0]
        self._cookies = None
        self._headers = GanmaRqHeaders(self._manga_alias)

    def manga_info(self):
        rq = requests.get(url='https://ganma.jp/api/2.0/magazines/' +
                          self._manga_alias,
                          headers=self._headers.asmangainfo(),
                          cookies=self._cookies)
        if rq.status_code != 200:
            logger.warning(str(rq.status_code) + ' ' + rq.text)
            return -1
        try:
            resp = rq.json()
        except ValueError as e:
            logger.error(
                "The response body does not contain valid json!\nstatus:{}, text:{}"
                .format(rq.status_code, rq.text))
            raise e
        mangainfo = {}
        if 'success' in resp and resp['success']:
            mangainfo['items'] = resp['root']['items']
            mangainfo['index_id'] = [x['id'] for x in mangainfo['items']]
            mangainfo['index_title'] = [
                x['title'] for x in mangainfo['items']
            ]

        return mangainfo

    def login(self, mail, passwd):
        if type(mail) != str or type(passwd) != str:
            logger.error("帳號（email）或密碼非法 type：{}，{}".format(
                type(mail), type(passwd))),
            raise ValueError("帳號（email）或密碼非法 type：{}，{}".format(
                type(mail), type(passwd)))
        rq = requests.post(url='https://ganma.jp/api/1.0/session?mode=mail',
                           json={
                               "password": passwd,
                               "mail": mail
                           },
                           headers=self._headers)

        try:
            resp = rq.json()
        except ValueError as e:
            logger.error(
                "The response body does not contain valid json!\nstatus:{}, text:{}"
                .format(rq.status_code, rq.text))
            raise e

        if 'success' in resp:
            if resp['success']:
                if 'PLAY_SESSION' in rq.cookies:
                    self._cookies = {
                        'PLAY_SESSION': rq.cookies['PLAY_SESSION']
                    }
                    return 0
            else:
                raise ValueError("帳號（email）或密碼錯誤")
        else:
            logger.error("The key 'success' does not exist! text:{}".format(
                rq.text))
            raise ValueError("The key 'success' does not exist!")

    def logout(self):
        rq = requests.delete(url="https://ganma.jp/api/1.0/session",
                             headers=self._headers.aslogout(),
                             cookies=self._cookies)
        self._cookies = None
        if rq.status_code == 200:
            return 0
        else:
            return -1

    @staticmethod
    def downld_one(url, fpath):
        rq = requests.get(url=url, headers=RqHeaders())
        if rq.status_code != 200:
            raise ValueError(url)
        else:
            with open(fpath, "wb") as fp:
                fp.write(rq.content)
        return 0

    def downloader(self):
        manga_info = self.manga_info()

        if self._param[1] and self._param[1] in manga_info["index_id"]:
            indx = manga_info["index_id"].index(self._param[1])
        else:
            raise ValueError("当前一话不存在或需要登录".format())

        dir = "./漫畫/" + \
            manga_info["items"][indx]["series"]["title"] + \
            "/" + manga_info["items"][indx]["title"]
        cc_mkdir(dir, model=1)

        progress_bar = ProgressBar(
            len(manga_info["items"][indx]["page"]["files"]))

        downld_gen = DownldGen(manga_info["items"][indx])

        with ThreadPoolExecutor(max_workers=8) as executor:
            count = 0
            for x in executor.map(self.downld_one, downld_gen.img_url_g,
                                  downld_gen.file_path_g):
                count += 1
                progress_bar.show(count)

        if manga_info["items"][indx]["afterwordImage"]["url"]:
            print("下載後記圖片！", end=" ")
            rq = requests.get(
                url=manga_info["items"][indx]["afterwordImage"]["url"], headers=RqHeaders())
            if rq.status_code != 200:
                logger.error("Error, afterword image: " +
                             manga_info["items"][indx]["afterwordImage"]["url"])
                raise ValueError(manga_info["items"]
                                 [indx]["afterwordImage"]["url"])
            with open(dir + "/afterword.jpg", "wb") as fp:
                fp.write(rq.content)
            print("成功！")
        else:
            print("No afterword image found!")



import copy
import json
import logging
import math
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO

import requests
from PIL import Image
from selenium import webdriver
from selenium.webdriver.support.wait import WebDriverWait

from .utils import ComicLinkInfo, ProgressBar, RqHeaders, SiteReaderLoad, cc_mkdir, draw_image

logger = logging.getLogger("comic-action")

WAIT_TIME = 60


class proc_img_co:
    def __init__(self, width, height):
        self.DIVIDE_NUM = 4
        self.MULTIPLE = 8
        self.width = width
        self.height = height
        self.cell_width = math.floor(
            self.width / (self.DIVIDE_NUM * self.MULTIPLE)) * self.MULTIPLE
        self.cell_height = math.floor(
            self.height / (self.DIVIDE_NUM * self.MULTIPLE)) * self.MULTIPLE

    def n21(self, img0) -> Image.Image:
        img_copy = copy.deepcopy(img0)
        for n in range(self.DIVIDE_NUM * self.DIVIDE_NUM):
            src_x = n % self.DIVIDE_NUM * self.cell_width
            src_y = math.floor(n / self.DIVIDE_NUM) * self.cell_height
            i = n % self.DIVIDE_NUM * self.DIVIDE_NUM + \
                math.floor(n / self.DIVIDE_NUM)
            t_x = i % self.DIVIDE_NUM * self.cell_width
            t_y = math.floor(i / self.DIVIDE_NUM) * self.cell_height
            draw_image(img0, img_copy, src_x, src_y, self.cell_width,
                       self.cell_height, t_x, t_y)
        return img_copy


@SiteReaderLoad.register('comic_action')
class ComicAction(object):
    def __init__(self, linkinfo: ComicLinkInfo, driver: webdriver.Chrome):
        super().__init__()
        self._linkinfo = linkinfo
        self._driver = driver

    @staticmethod
    def get_comic_json(linkinfo: ComicLinkInfo, driver = None):
        r"""
        """
        comic_json = {
            "title": None,
            "subtitle": None,
            "pages": [],
        }
        if linkinfo.param[1] == 1 and driver:
            driver.get(linkinfo.url)
            elem = WebDriverWait(driver, WAIT_TIME, 0.5).until(
                lambda x: x.find_element_by_id("episode-json"),
                message="無法定位元素 episode-json" + ", 獲取json對象數據失敗")

            json_dataValue = elem.get_attribute("data-value")
            json_dataValue = json.loads(json_dataValue)
        elif linkinfo.param[1] == 0:
            rq = requests.get(linkinfo.url+".json", headers = RqHeaders())
            if rq.status_code != 200:
                raise ValueError(linkinfo.url+".json")
            json_dataValue = rq.json()
        else:
            raise ValueError("linkinfo.param[1] not 1 or 0, :"+linkinfo.site_name)
        comic_json["subtitle"] = json_dataValue["readableProduct"]["title"].replace("?", "？")
        comic_json["title"] = json_dataValue["readableProduct"]["series"]["title"].replace("?", "？")
        for page in json_dataValue["readableProduct"]["pageStructure"]["pages"]:
            if "src" in page:
                comic_json["pages"].append(page)
        return comic_json

    @staticmethod
    def gen_url(comic_json: dict):
        for x in comic_json["pages"]:
            yield x["src"]

    @staticmethod
    def gen_fpth(comic_json: dict):
        bpth = "./漫畫/" + \
            "/".join((comic_json["title"], comic_json["subtitle"]))
        count = 0
        for x in range(len(comic_json["pages"])):
            count += 1
            yield [bpth, "{}.png".format(count)]

    @staticmethod
    def downld_one(url, fpth, cookies=None):
        r"""
        :fpth: [basepath, fname]
        """
        rq = requests.get(url, headers=RqHeaders())
        if rq.status_code != 200:
            raise ValueError(url)
        content = rq.content
        img0 = Image.open(BytesIO(content))

        # 源图像
        img0.save(fpth[0] + "/source/" + fpth[1])

        # 复原
        proc = proc_img_co(img0.width, img0.height)
        proc.n21(img0=img0).save(fpth[0] + "/target/" + fpth[1])

    def downloader(self):
        # https://<domain: comic-action.com ...>/episode/13933686331648942300
        comic_json = self.get_comic_json(self._linkinfo, self._driver)
        total_pages = len(comic_json["pages"])
        show_bar = ProgressBar(total_pages)
        cc_mkdir("./漫畫/" + \
            "/".join((comic_json["title"], comic_json["subtitle"])))
        with ThreadPoolExecutor(max_workers=4) as executor:
            count = 0
            for x in executor.map(self.downld_one,
                                  self.gen_url(comic_json),
                                  self.gen_fpth(comic_json)):
                count += 1
                show_bar.show(count)


# if __name__ == "__main__":
#     url = "https://comic-action.com/episode/13933686331648942300"
#     driver = webdriver.Chrome()
#     ComicAction().downloader(driver, ComicLinkInfo(url))



import json
import logging
import re
from concurrent.futures import ThreadPoolExecutor

import requests

from .utils import ComicLinkInfo, ProgressBar, RqHeaders, SiteReaderLoad, cc_mkdir

logger = logging.getLogger(__name__)


# 生成图像解码key
def gen_key_8(url_img_hash: str):
    key_8 = []
    for x in range(0, 16, 2):
        h = url_img_hash[x]+url_img_hash[x+1]
        d = int(h, 16)
        key_8.append(d)
    return key_8


def xor_img(arr_img: list, url_img_hash):
    arr_key = gen_key_8(url_img_hash)
    img_ = []
    for a in range(len(arr_img)):
        img_.append(arr_img[a] ^ arr_key[a % 8])
    return img_


def write2jpg(img_, fpath):
    with open(fpath, 'wb') as fp:
        for x in img_:
            fp.write((x).to_bytes(length=1, byteorder='big'))

@SiteReaderLoad.register('comic_walker')
class ComicWalker(object):
    def __init__(self, link_info: ComicLinkInfo, driver=None):
        super().__init__()
        self._link_info = link_info
        self._driver = driver

    @staticmethod
    def downld_one(item, fpath):
        url = item['meta']['source_url']
        rr = requests.get(url, headers=RqHeaders())
        if rr.status_code != 200:
            raise ValueError(item['meta']['source_url'])
        write2jpg(xor_img(
            rr.content, item['meta']['drm_hash']), fpath)

    def downloader(self):
        comic_cid = self._link_info.param[0][0]
        rq = requests.get(self._link_info.url, headers=RqHeaders())
        if rq.status_code != 200:
            raise ValueError(self._link_info.url)
        comic_info = json.loads(
            re.search("<script>\sdataLayer[\s\S]*?(\[.*\])[\s\S]*?</script>", rq.text).group(1))[0]
        base_fpath = "./漫畫/" + \
            "/".join([comic_info["content_title"],
                      comic_info["episode_title"]])
        # https://comicwalker-api.nicomanga.jp/api/v1/comicwalker/
        # https://ssl.seiga.nicovideo.jp/api/v1/comicwalker/episodes/
        url_json_comic = 'https://comicwalker-api.nicomanga.jp/api/v1/comicwalker/episodes/' + \
            comic_cid+'/frames'
        r = requests.get(url=url_json_comic, headers=RqHeaders())
        r_json = r.json()
        if cc_mkdir(base_fpath, model=1) != 0:
            return -1
        show_bar = ProgressBar(len(r_json['data']['result']))
        items = [x for x in r_json['data']['result']]
        fpth_l = [base_fpath+"/"+str(x)+".jpg" for x in range(1, len(r_json['data']['result'])+1)]
        with ThreadPoolExecutor(max_workers=4) as executor:
            count = 0
            for x in executor.map(self.downld_one, items, fpth_l):
                count += 1
                show_bar.show(count)

# if __name__ == "__main__":
#     'https://comic-walker.com/viewer/?tw=2&dlcl=ja&cid=KDCW_MF00000051010034_68'



import setuptools

with open("./src/README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ccdl", 
    version="0.1.11",
    author="vircoys",
    author_email="vircoys@gmail.com",
    description="Online comic downloader",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vircoys/ccdl",
    license="unlicense",
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "lxml",
        "numpy",
        "Pillow",
        "requests",
        "selenium",
        "pyexecjs"
    ]
)